package meta.system.bop;

import java.util.Vector;
import java.util.Collection;
import java.lang.reflect.Field;
import org.aspectj.lang.Signature;
import org.aspectj.lang.reflect.FieldSignature;
import java.sql.ResultSet;
import java.sql.SQLException;
import meta.system.managers.StaticManagerHandles;
import meta.system.registry.Registable;

/**
 * Defines the functionality of {@link meta.system.bop.BOPPersistable}. Manages the addition of Fields to {@link meta.system.bop.BOPPersistable#modified}. 
 * Defines a simple pointcut that joins on the 'set' of any field defined in a {@link meta.system.bop.BOPPersistable} extention.
 */
public aspect BOPPersistableAugment {	
	
	/**
	 * Used to store Fields that are modified during an instances life time. This collection is added to in the
	 * aspect {@link meta.system.bop.BOPPersistableAugment}.
	 */
	public Collection<Field> BOPPersistable.modified = new Vector<Field>(5, 2);
	
	/**
	 * Used to flag weather an instance should be inserted or updated, defaults to false.
	 */
	public boolean BOPPersistable.original = false;
	
	/**
	 * The point cut definition. Joins on <code>set(* HOPPersistable+.*)</code>, also gets a handle on the instance the
	 * set is being called on.
	 */
	pointcut sets(BOPPersistable entity) : set(@PersistField * *) &&
						!set(Vector<Field> BOPPersistable.modified) && //don't wanna record those!
						this(entity); 
	
	/**
	 * Advice for the pointcut. Simply extracts the FieldSignature from the static part of the join point and, if not
	 * already contianed within, adds it to the {@link meta.system.bop.BOPPersistable#modified} of the instance in question.
	 */
	before(BOPPersistable entity) : sets(entity){
		Signature sig = thisJoinPointStaticPart.getSignature();
		//attempt cast to field type
		FieldSignature fSig = (FieldSignature)sig;
		
		Field field = fSig.getField();
		
		if(!entity.modified.contains(fSig.getField()))//TODO test for @Persist
		entity.modified.add(field);
	}
	
	/**
	 * Returns the Vector containing Fields that have been modified since this instances creation.
	 * @return the Vector.
	 */
	public Collection BOPPersistable.getModified(){
		return this.modified;
	}
	
	/**
	 * Clears the contents of {@link #modified}.
	 */
	public void BOPPersistable.clearModified(){
		this.modified.clear();
	}
	
	
	/*
	 * Registable Bopper advice
	 */
	
	pointcut deregistered(BOPPersistable entity) : call(public void Registable+.deregistered()) && target(entity); 
	
	/**
	 * Causes this to be to be sent to {@link meta.system.bop.BasicObjectPersistence} for Database storage. If 
	 * {@link meta.system.bop.BOPPersistableAugment.original} is true then this is sent to 
	 * {@link meta.system.bop.BasicObjectPersistence#insertEntity} if false then this is sent to
	 * {@link meta.system.bop.BasicObjectPersistence#updateEntity}.
	 */
	after(BOPPersistable entity) throws Exception : deregistered(entity){
		if(entity.isOriginal()){
			StaticManagerHandles.TheBigBopper.insertEntity((BOPPersistable)entity);
		}else{
			StaticManagerHandles.TheBigBopper.updateEntity((BOPPersistable)entity);
		}		
	}
	
	
	public boolean BOPPersistable.isOriginal(){
		return this.original;
	}
	
	public void BOPPersistable.setOriginal(boolean orig){
		this.original = orig;
	}
	
	
	
	
	/*
	 * Collection deletion support.
	 */
	
	/*This does actually provide the functionality thameta.services.fraudguard.vyc.VYCModuleManager
t we want, BUT it will create a conditional join on every
	 * remove on any collection, only those that remove an ArrayPersistable will continue to the advice, but still..
	 * Don't think its acceptable. So have implemented a bare face remove method in the persistence manager.
	 * Rubbish i have been defeated. Need a kind of 'memberOf' AOP construct, so could specifie only to match on Collections that
	 * were fields in PersistableEntitys, but one does not exist, yet... um maybe it does..
	 * Cld maybe do this with a combination of pointcuts...
	 * no time to solve but cld be interesting.
	pointcut remove(HBOPArrayPersistable deleted) : call(public * java.util.Collection+.remove (..)) &&
											//		memberOf(HBOPPersistable+)&&
												//	target(java.util.Collection+)&&
												//	within(HBOPPersistable+)&&
													args(HBOPArrayPersistable)&& args(deleted);
	
	before(HBOPArrayPersistable deleted) : remove(deleted){
		HomeBrewObjPersistents.deleteElement(deleted);
	}
	
	*/
}
