package meta.system.registry;

import meta.system.registry.Cacheable;
import javolution.util.FastMap;
import meta.system.registry.Registable;
import org.apache.log4j.Logger;


/**
 * This aspect augments the Cachable interface to provide the logic for aging the entry.
 * Using a general joinpoint any method thats is called on a cachable entity causes the timer
 * to be reset, when the timer is found (by the remove oldies method called by the ChaceCullers) to be
 * to old the entity is removed from cahce.
 * 
 * Using this technique any added methods to any object wanting to be cached will have this behaivour.
 *
 */
public aspect CachableAugment {
	
	static final Logger log = Logger.getLogger("Augment");
	
	//records the time when a method was last used on the implementing object instance
	private long Cacheable.lastUsed = 0;
	//set to true when the entry is to be removed from the cache, 
	//used at the mo to stop cache cullers continuing thier existence after their cache has died 
	public boolean Cacheable.old = false;
	
	private Integer Cacheable.lifeTime = 5000;
	
	/**
	 * Pointcut for the olderThan method, we don't want this resetting the timer
	 */
	pointcut olderThanCalled() : execution(boolean Cacheable.olderThan( .. ) );
	/**
	 * Pointcut for the getLifeTime method, we don't want this resetting the timer
	 */
	pointcut getLifeTimeCalled() : execution(Integer Cacheable.getLifeTime() );
	/**
	 * Pointcut for the getLifeTime method, we don't want this resetting the timer
	 */
	pointcut getContentsCalled() : execution(FastMap Registable+.getContents() );

	/**
	 * Pointcut for the run method, we don't want this resetting the timer
	 */
	pointcut registryRun() : execution(void CachingRegistry.run());
	
	pointcut registryFlush() : execution(void CachingRegistry.flush());
	
	/**
	 * Pointcut for the methods added by this aspect to chacheable
	 */
	pointcut cachSetOld() : execution(void Cacheable+.setOld( .. ));
	
	
	pointcut cachIsOld() : execution(boolean Cacheable+.isOld( .. ));
	/**
	 * Pointcut for the methods added by this aspect to chacheable
	 */
	pointcut deregister() : execution(void Cacheable+.deregistered( .. ));
	
	
	
	/**
	 * Should match on all method of any class that implements the flush()Cacheable interface. Caveats, excludes any static methods
	 * of implementing classes and excludes the CachingRegistry.removeOldies() method. If we did'nt exclude this
	 * then no chachingRegistry would wld ever be considered to old because the cacheCuller would be resetting its clock all the time.  
	 */
	pointcut allCacheableMethods(Cacheable entity) : execution(public * Cacheable+.*( .. )) && !execution(public static * Cacheable+.* ( .. )) 
													 && this(entity) && !execution(public * CachingRegistry.removeOldies());
	
	before(Cacheable entity) : allCacheableMethods(entity) && ! olderThanCalled() && !registryRun() && !registryFlush() 
															&& !cachSetOld() && !cachIsOld() && !getLifeTimeCalled() && !getContentsCalled()
															&& !deregister(){		
		//below is a debug output thats very helpful when caching weridness happens.
		log.debug("Updating timer for "+thisJoinPoint.getThis().getClass()+" cause "+thisJoinPointStaticPart.getSignature()+" has been called.");
	
		entity.lastUsed = System.currentTimeMillis();
	}

	/**
	 * Returns true when the age (ie the amount of time since a method on this object was called)
	 * of the object is less than the passed value.
	 * @param time
	 * @return
	 */
	public boolean Cacheable.olderThan(long aTime){
		if(aTime > lastUsed){
			return true;
		}
		return false;
	}
	
	public boolean Cacheable.setOld(boolean _old){
		old = _old;
		return true;
	}
	
	public boolean Cacheable.isOld(){
		return old;
	}
	
	public Integer Cacheable.getLifeTime(){
		return lifeTime; 
	}
	/**
	 * Sets the amount of time this entity will live for in the cache, set to null to gain immortality.
	 */
	public void Cacheable.setLifeTime(Integer newLife){
		lifeTime = newLife;
	}
	
}
